"""Primitive Gaussian type orbitals"""

from typing import Optional

import equinox as eqx
import jax.numpy as jnp
import numpy as np
from jax import jit
from jax.scipy.special import gammaln

from mess.types import Float3, FloatN, FloatNx3, Int3, asintarray, asfparray


class Primitive(eqx.Module):
    center: Float3 = eqx.field(converter=asfparray, default=(0.0, 0.0, 0.0))
    alpha: float = eqx.field(converter=asfparray, default=1.0)
    lmn: Int3 = eqx.field(converter=asintarray, default=(0, 0, 0))
    norm: Optional[float] = None

    def __post_init__(self):
        if self.norm is None:
            self.norm = normalize(self.lmn, self.alpha)

    def __check_init__(self):
        names = ["center", "alpha", "lmn", "norm"]
        shapes = [(3,), (), (3,), ()]
        dtypes = [jnp.floating, jnp.floating, jnp.integer, jnp.floating]

        for name, shape, dtype in zip(names, shapes, dtypes):
            value = getattr(self, name)
            if value.shape != shape or not jnp.issubdtype(value, dtype):
                raise ValueError(
                    f"Invalid value for {name}.\n"
                    f"Expecting {dtype} array with shape {shape}. "
                    f"Got {value.dtype} with shape {value.shape}"
                )

    @property
    def angular_momentum(self) -> int:
        return np.sum(self.lmn)

    def __call__(self, pos: FloatNx3) -> FloatN:
        return eval_primitive(self, pos)

    def __hash__(self) -> int:
        values = []
        for k, v in vars(self).items():
            if k.startswith("__") or v is None:
                continue

            values.append(v.tobytes())

        return hash(b"".join(values))


@jit
def normalize(lmn: Int3, alpha: float) -> float:
    L = jnp.sum(lmn)
    N = ((1 / 2) / alpha) ** (L + 3 / 2)
    N *= jnp.exp(jnp.sum(gammaln(lmn + 1 / 2)))
    return N**-0.5


def product(a: Primitive, b: Primitive) -> Primitive:
    alpha = a.alpha + b.alpha
    center = (a.alpha * a.center + b.alpha * b.center) / alpha
    lmn = a.lmn + b.lmn
    c = a.norm * b.norm
    Rab = a.center - b.center
    c *= jnp.exp(-a.alpha * b.alpha / alpha * jnp.inner(Rab, Rab))
    return Primitive(center=center, alpha=alpha, lmn=lmn, norm=c)


def eval_primitive(p: Primitive, pos: FloatNx3) -> FloatN:
    pos = jnp.atleast_2d(pos)
    assert pos.ndim == 2 and pos.shape[1] == 3, "pos must have shape [N,3]"
    pos_translated = pos[:, jnp.newaxis] - p.center
    v = p.norm * jnp.exp(-p.alpha * jnp.sum(pos_translated**2, axis=-1))
    v *= jnp.prod(pos_translated**p.lmn, axis=-1)
    return jnp.squeeze(v)



import numpy as np
from numpy.testing import assert_allclose

from mess.basis import basisset
from mess.interop import to_pyscf
from mess.autograd_integrals import (
    grad_overlap_basis,
    grad_kinetic_basis,
    grad_nuclear_basis,
)
from mess.structure import molecule


def test_nuclear_gradients():
    basis_name = "sto-3g"
    h2 = molecule("h2")
    scfmol = to_pyscf(h2, basis_name)
    basis = basisset(h2, basis_name)

    actual = grad_overlap_basis(basis)
    expect = scfmol.intor("int1e_ipovlp_cart", comp=3)
    assert_allclose(actual, expect, atol=1e-6)

    actual = grad_kinetic_basis(basis)
    expect = scfmol.intor("int1e_ipkin_cart", comp=3)
    assert_allclose(actual, expect, atol=1e-6)

    # TODO: investigate possible inconsistency in libcint outputs?
    actual = grad_nuclear_basis(basis)
    expect = scfmol.intor("int1e_ipnuc_cart", comp=3)
    expect = -np.moveaxis(expect, 1, 2)
    assert_allclose(actual, expect, atol=1e-5)



import jax
import pytest

from mess.basis import basisset
from mess.hamiltonian import Hamiltonian, minimise
from mess.zeropad_integrals import overlap_basis_zeropad
from mess.integrals import (
    eri_basis_sparse,
    kinetic_basis,
    nuclear_basis,
    overlap_basis,
)
from mess.structure import molecule
from mess.interop import from_pyquante
from mess.package_utils import has_package
from conftest import is_mem_limited


@pytest.mark.parametrize("func", [overlap_basis, overlap_basis_zeropad, kinetic_basis])
@pytest.mark.skipif(
    not has_package("pyquante2"), reason="Missing Optional Dependency: pyquante2"
)
def test_benzene(func, benchmark):
    mol = from_pyquante("c6h6")
    basis = basisset(mol, "def2-TZVPPD")
    basis = jax.device_put(basis)

    def harness():
        return func(basis).block_until_ready()

    benchmark(harness)


@pytest.mark.parametrize("mol_name", ["h2", "water"])
@pytest.mark.parametrize(
    "func", [overlap_basis, kinetic_basis, nuclear_basis, eri_basis_sparse]
)
def test_integrals(mol_name, func, benchmark):
    mol = molecule(mol_name)
    basis = basisset(mol)
    basis = jax.device_put(basis)

    def harness():
        return func(basis).block_until_ready()

    benchmark(harness)


@pytest.mark.parametrize("mol_name", ["h2", "water"])
@pytest.mark.skipif(is_mem_limited(), reason="Not enough host memory!")
def test_minimise_ks(benchmark, mol_name):
    # TODO: investigate test failure with cpu backend and float32
    from jax.experimental import enable_x64

    with enable_x64(True):
        mol = molecule(mol_name)
        basis = basisset(mol, "6-31g")
        H = Hamiltonian(basis)
        H = jax.device_put(H)

        def harness():
            E, C, _ = minimise(H)
            return E.block_until_ready(), C.block_until_ready()

        benchmark(harness)



r"""Orthonormal transformation.

Evaluates the transformation matrix :math:`X` that satisfies

.. math:: \mathbf{X}^T \mathbf{S} \mathbf{X} = \mathbb{I}

where :math:`\mathbf{S}` is the overlap matrix of the non-orthonormal basis and
:math:`\mathbb{I}` is the identity matrix.

This module implements a few commonly used orthonormalisation transforms.
"""

import jax.numpy as jnp
import jax.numpy.linalg as jnl

from mess.types import FloatNxN


def canonical(S: FloatNxN) -> FloatNxN:
    r"""Canonical orthonormal transformation

    .. math:: \mathbf{X} = \mathbf{U} \mathbf{s}^{-1/2}

    where :math:`\mathbf{U}` and :math:`\mathbf{s}` are the eigenvectors and
    eigenvalues of the overlap matrix :math:`\mathbf{S}`.

    Args:
        S (FloatNxN): overlap matrix for the non-orthonormal basis.

    Returns:
        FloatNxN: canonical orthonormal transformation matrix
    """
    s, U = jnl.eigh(S)
    s = jnp.diag(jnp.power(s, -0.5))
    return U @ s


def symmetric(S: FloatNxN) -> FloatNxN:
    r"""Symmetric orthonormal transformation

    .. math:: \mathbf{X} = \mathbf{U} \mathbf{s}^{-1/2} \mathbf{U}^T

    where :math:`\mathbf{U}` and :math:`\mathbf{s}` are the eigenvectors and
    eigenvalues of the overlap matrix :math:`\mathbf{S}`.

    Args:
        S (FloatNxN): overlap matrix for the non-orthonormal basis.

    Returns:
        FloatNxN: symmetric orthonormal transformation matrix
    """
    s, U = jnl.eigh(S)
    s = jnp.diag(jnp.power(s, -0.5))
    return U @ s @ U.T


def cholesky(S: FloatNxN) -> FloatNxN:
    r"""Cholesky orthonormal transformation

    .. math:: \mathbf{X} = (\mathbf{L}^{-1})^T

    where :math:`\mathbf{L}` is the lower triangular matrix that satisfies the Cholesky
    decomposition of the overlap matrix :math:`\mathbf{S}`.

    Args:
        S (FloatNxN): overlap matrix for the non-orthonormal basis.

    Returns:
        FloatNxN: cholesky orthonormal transformation matrix
    """
    L = jnl.cholesky(S)
    return jnl.inv(L).T



"""
JAX implementation for integrals over Gaussian basis functions.

Based upon the closed-form expressions derived in

    Taketa, H., Huzinaga, S., & O-ohata, K. (1966). Gaussian-expansion methods for
    molecular integrals. Journal of the physical society of Japan, 21(11), 2313-2324.
    <https://doi.org/10.1143/JPSJ.21.2313>

Hereafter referred to as the "THO paper"

Related work:

[1] Augspurger JD, Dykstra CE. General quantum mechanical operators. An
    open-ended approach for one-electron integrals with Gaussian bases. Journal of
    computational chemistry. 1990 Jan;11(1):105-11.
    <https://doi.org/10.1002/jcc.540110113>
"""

from dataclasses import asdict
from functools import partial
from itertools import product as cartesian_product
from more_itertools import batched
from typing import Callable

import jax.numpy as jnp
import numpy as np
from jax import jit, tree, vmap
from jax.ops import segment_sum

from mess.basis import Basis, basis_iter
from mess.primitive import Primitive, product
from mess.special import (
    binom,
    binom_factor,
    factorial,
    factorial2,
    gammanu,
    allpairs_indices,
)
from mess.types import Float3, FloatNxN
from mess.units import LMAX

BinaryPrimitiveOp = Callable[[Primitive, Primitive], float]


@partial(jit, static_argnums=(0, 1))
def integrate_dense(basis: Basis, primitive_op: BinaryPrimitiveOp) -> FloatNxN:
    (ii, cl, lhs), (jj, cr, rhs) = basis_iter(basis)
    aij = cl * cr * vmap(primitive_op)(lhs, rhs)
    A = jnp.zeros((basis.num_primitives, basis.num_primitives))
    A = A.at[ii, jj].set(aij)
    A = A + A.T - jnp.diag(jnp.diag(A))
    index = basis.orbital_index.reshape(1, basis.num_primitives)
    out = segment_sum(A, index, num_segments=basis.num_orbitals)
    out = segment_sum(out.T, index, num_segments=basis.num_orbitals)
    return out


@partial(jit, static_argnums=(0, 1))
def integrate_sparse(basis: Basis, primitive_op: BinaryPrimitiveOp) -> FloatNxN:
    offset = [0] + [o.num_primitives for o in basis.orbitals]
    offset = np.cumsum(offset)
    ii, jj = allpairs_indices(basis.num_orbitals)
    indices = []
    batch = []

    for count, idx in enumerate(zip(ii, jj)):
        mesh = [range(offset[i], offset[i + 1]) for i in idx]
        indices += list(cartesian_product(*mesh))
        batch += [count] * (len(indices) - len(batch))

    indices = np.array(indices, dtype=np.int32).T
    batch = np.array(batch, dtype=np.int32)
    cij = jnp.stack([jnp.take(basis.coefficients, idx) for idx in indices]).prod(axis=0)
    pij = [
        tree.map(lambda x: jnp.take(x, idx, axis=0), basis.primitives)
        for idx in indices
    ]
    aij = segment_sum(cij * vmap(primitive_op)(*pij), batch, num_segments=count + 1)

    A = jnp.zeros_like(aij, shape=(basis.num_orbitals, basis.num_orbitals))
    A = A.at[ii, jj].set(aij)
    A = A + A.T - jnp.diag(jnp.diag(A))
    return A


integrate = integrate_dense


def _overlap_primitives(a: Primitive, b: Primitive) -> float:
    @vmap
    def overlap_axis(i: int, j: int, a: float, b: float) -> float:
        idx = [(s, t) for s in range(LMAX + 1) for t in range(2 * s + 1)]
        s, t = jnp.array(idx, dtype=jnp.uint32).T
        out = binom(i, 2 * s - t) * binom(j, t)
        out *= a ** jnp.maximum(i - (2 * s - t), 0) * b ** jnp.maximum(j - t, 0)
        out *= factorial2(2 * s - 1) / (2 * p.alpha) ** s

        mask = (2 * s - i <= t) & (t <= j)
        out = jnp.where(mask, out, 0)
        return jnp.sum(out)

    p = product(a, b)
    pa = p.center - a.center
    pb = p.center - b.center
    out = jnp.power(jnp.pi / p.alpha, 1.5) * p.norm
    out *= jnp.prod(overlap_axis(a.lmn, b.lmn, pa, pb))
    return out


def overlap_basis(basis: Basis) -> FloatNxN:
    return integrate(basis, _overlap_primitives)


def _kinetic_primitives(a: Primitive, b: Primitive) -> float:
    t0 = b.alpha * (2 * jnp.sum(b.lmn) + 3) * _overlap_primitives(a, b)

    def offset_qn(ax: int, offset: int):
        lmn = b.lmn.at[ax].add(offset)
        return Primitive(**{**asdict(b), "lmn": lmn})

    axes = jnp.arange(3)
    b1 = vmap(offset_qn, (0, None))(axes, 2)
    t1 = jnp.sum(vmap(_overlap_primitives, (None, 0))(a, b1))

    b2 = vmap(offset_qn, (0, None))(axes, -2)
    t2 = jnp.sum(b.lmn * (b.lmn - 1) * vmap(_overlap_primitives, (None, 0))(a, b2))
    return t0 - 2.0 * b.alpha**2 * t1 - 0.5 * t2


def kinetic_basis(b: Basis) -> FloatNxN:
    return integrate(b, _kinetic_primitives)


def build_gindex():
    vals = [
        (i, r, u)
        for i in range(LMAX + 1)
        for r in range(i // 2 + 1)
        for u in range((i - 2 * r) // 2 + 1)
    ]
    i, r, u = jnp.array(vals).T
    return i, r, u


gindex = build_gindex()


def _nuclear_primitives(a: Primitive, b: Primitive, c: Float3):
    p = product(a, b)
    pa = p.center - a.center
    pb = p.center - b.center
    pc = p.center - c
    epsilon = 1.0 / (4.0 * p.alpha)

    @vmap
    def g_term(l1, l2, pa, pb, cp):
        i, r, u = gindex
        index = i - 2 * r - u
        g = (
            jnp.power(-1, i + u)
            * jnp.take(binom_factor(l1, l2, pa, pb), i)
            * factorial(i)
            * jnp.power(cp, index - u)
            * jnp.power(epsilon, r + u)
        ) / (factorial(r) * factorial(u) * factorial(index - u))

        g = jnp.where(index <= l1 + l2, g, 0.0)
        return segment_sum(g, index, num_segments=LMAX + 1)

    Gi, Gj, Gk = g_term(a.lmn, b.lmn, pa, pb, pc)
    ids = jnp.arange(3 * LMAX + 1)
    ijk = jnp.arange(LMAX + 1)
    nu = (
        ijk[:, jnp.newaxis, jnp.newaxis]
        + ijk[jnp.newaxis, :, jnp.newaxis]
        + ijk[jnp.newaxis, jnp.newaxis, :]
    )

    W = (
        Gi[:, jnp.newaxis, jnp.newaxis]
        * Gj[jnp.newaxis, :, jnp.newaxis]
        * Gk[jnp.newaxis, jnp.newaxis, :]
        * jnp.take(gammanu(ids, p.alpha * jnp.inner(pc, pc)), nu)
    )

    return -2.0 * jnp.pi / p.alpha * p.norm * jnp.sum(W)


overlap_primitives = jit(_overlap_primitives)
kinetic_primitives = jit(_kinetic_primitives)
nuclear_primitives = jit(_nuclear_primitives)


@partial(jit, static_argnums=0)
def nuclear_basis(basis: Basis):
    def n(atomic_number, position):
        def op(pi, pj):
            return atomic_number * _nuclear_primitives(pi, pj, position)

        return integrate(basis, op)

    return vmap(n)(basis.structure.atomic_number, basis.structure.position)


def build_cindex():
    vals = [
        (i1, i2, r1, r2, u)
        for i1 in range(2 * LMAX + 1)
        for i2 in range(2 * LMAX + 1)
        for r1 in range(i1 // 2 + 1)
        for r2 in range(i2 // 2 + 1)
        for u in range((i1 + i2) // 2 - r1 - r2 + 1)
    ]
    i1, i2, r1, r2, u = jnp.array(vals).T
    return i1, i2, r1, r2, u


cindex = build_cindex()


def _eri_primitives(a: Primitive, b: Primitive, c: Primitive, d: Primitive) -> float:
    p = product(a, b)
    q = product(c, d)
    pa = p.center - a.center
    pb = p.center - b.center
    qc = q.center - c.center
    qd = q.center - d.center
    qp = q.center - p.center
    delta = 1 / (4.0 * p.alpha) + 1 / (4.0 * q.alpha)

    def H(l1, l2, a, b, i, r, gamma):
        # Note this should match THO Eq 3.5 but that seems to incorrectly show a
        # 1/(4 gamma)^(i- 2r) term which is inconsistent with Eq 2.22.
        # Using (4 gamma)^(r - i) matches the reported expressions for H_L
        u = factorial(i) * jnp.take(binom_factor(l1, l2, a, b, 2 * LMAX), i)
        v = factorial(r) * factorial(i - 2 * r) * (4 * gamma) ** (i - r)
        return u / v

    @vmap
    def c_term(la, lb, lc, ld, pa, pb, qc, qd, qp):
        # THO Eq 2.22 and 3.4
        i1, i2, r1, r2, u = cindex
        h = H(la, lb, pa, pb, i1, r1, p.alpha) * H(lc, ld, qc, qd, i2, r2, q.alpha)
        index = i1 + i2 - 2 * (r1 + r2) - u
        x = (-1) ** (i2 + u) * factorial(index + u) * qp ** (index - u)
        y = factorial(u) * factorial(index - u) * delta**index
        c = h * x / y

        mask = (i1 <= (la + lb)) & (i2 <= (lc + ld))
        c = jnp.where(mask, c, 0.0)
        return segment_sum(c, index, num_segments=4 * LMAX + 1)

    Ci, Cj, Ck = c_term(a.lmn, b.lmn, c.lmn, d.lmn, pa, pb, qc, qd, qp)

    ijk = jnp.arange(4 * LMAX + 1)
    nu = (
        ijk[:, jnp.newaxis, jnp.newaxis]
        + ijk[jnp.newaxis, :, jnp.newaxis]
        + ijk[jnp.newaxis, jnp.newaxis, :]
    )
    ids = jnp.arange(12 * LMAX + 1)

    W = (
        Ci[:, jnp.newaxis, jnp.newaxis]
        * Cj[jnp.newaxis, :, jnp.newaxis]
        * Ck[jnp.newaxis, jnp.newaxis, :]
        * jnp.take(gammanu(ids, jnp.inner(qp, qp) / (4.0 * delta)), nu)
    )

    return (
        2.0
        * jnp.pi**2
        / (p.alpha * q.alpha)
        * jnp.sqrt(jnp.pi / (p.alpha + q.alpha))
        * p.norm
        * q.norm
        * jnp.sum(W)
    )


eri_primitives = jit(_eri_primitives)
vmap_eri_primitives = jit(vmap(_eri_primitives))


def gen_ijkl(n: int):
    """Adapted from four-index transformations by S Wilson pg 257"""
    for idx in range(n):
        for jdx in range(idx + 1):
            for kdx in range(idx + 1):
                lmax = jdx if idx == kdx else kdx
                for ldx in range(lmax + 1):
                    yield idx, jdx, kdx, ldx


@partial(jit, static_argnums=0)
def eri_basis_sparse(b: Basis):
    indices = []
    batch = []
    offset = np.cumsum([o.num_primitives for o in b.orbitals])
    offset = np.insert(offset, 0, 0)

    for count, idx in enumerate(gen_ijkl(b.num_orbitals)):
        mesh = [range(offset[i], offset[i + 1]) for i in idx]
        indices += list(cartesian_product(*mesh))
        batch += [count] * (len(indices) - len(batch))

    indices = jnp.array(indices, dtype=jnp.int32).T
    batch = jnp.array(batch, dtype=jnp.int32)
    cijkl = jnp.stack([jnp.take(b.coefficients, idx) for idx in indices]).prod(axis=0)
    pijkl = [
        tree.map(lambda x: jnp.take(x, idx, axis=0), b.primitives) for idx in indices
    ]
    eris = cijkl * vmap_eri_primitives(*pijkl)
    return segment_sum(eris, batch, num_segments=count + 1)


def eri_basis(b: Basis):
    unique_eris = eri_basis_sparse_batched(b, batch_size=1024 * 1024)
    ii, jj, kk, ll = jnp.array(list(gen_ijkl(b.num_orbitals)), dtype=jnp.int32).T

    # Apply 8x permutation symmetry to build dense ERI from sparse ERI.
    eri_dense = jnp.empty_like(unique_eris, shape=(b.num_orbitals,) * 4)
    eri_dense = eri_dense.at[ii, jj, kk, ll].set(unique_eris)
    eri_dense = eri_dense.at[ii, jj, ll, kk].set(unique_eris)
    eri_dense = eri_dense.at[jj, ii, kk, ll].set(unique_eris)
    eri_dense = eri_dense.at[jj, ii, ll, kk].set(unique_eris)
    eri_dense = eri_dense.at[kk, ll, ii, jj].set(unique_eris)
    eri_dense = eri_dense.at[kk, ll, jj, ii].set(unique_eris)
    eri_dense = eri_dense.at[ll, kk, ii, jj].set(unique_eris)
    eri_dense = eri_dense.at[ll, kk, jj, ii].set(unique_eris)
    return eri_dense


def eri_basis_sparse_batched(basis: Basis, batch_size: int):
    batched_eris = []
    offset = np.cumsum([o.num_primitives for o in basis.orbitals])
    offset = np.insert(offset, 0, 0)

    for ijkl_batch in batched(gen_ijkl(basis.num_orbitals), batch_size):
        indices = []
        batch = []
        for count, idx in enumerate(ijkl_batch):
            mesh = [range(offset[i], offset[i + 1]) for i in idx]
            indices += list(cartesian_product(*mesh))
            batch += [count] * (len(indices) - len(batch))

        indices = jnp.array(indices, dtype=jnp.int32).T
        batch = jnp.array(batch, dtype=jnp.int32)
        cijkl = jnp.stack([jnp.take(basis.coefficients, idx) for idx in indices]).prod(
            axis=0
        )
        pijkl = [
            tree.map(lambda x: jnp.take(x, idx, axis=0), basis.primitives)
            for idx in indices
        ]
        eris = cijkl * vmap_eri_primitives(*pijkl)
        batched_eris += [segment_sum(eris, batch, num_segments=count + 1)]

    return jnp.hstack(batched_eris)



"""Discretised sampling of orbitals and charge density."""

from typing import Optional, Tuple, Union

import equinox as eqx
import jax.numpy as jnp
from pyscf import dft
from jax import vjp

from mess.basis import Basis
from mess.interop import to_pyscf
from mess.structure import Structure
from mess.types import FloatN, FloatNx3, FloatNxN, MeshAxes


class Mesh(eqx.Module):
    points: FloatNx3
    weights: Optional[FloatN] = None
    axes: Optional[MeshAxes] = None


def uniform_mesh(
    n: Union[int, Tuple] = 50, b: Union[float, Tuple] = 10.0, ndim: int = 3
) -> Mesh:
    if isinstance(n, int):
        n = (n,) * ndim

    if isinstance(b, float):
        b = (b,) * ndim

    if not isinstance(n, (tuple, list)):
        raise ValueError("Expected an integer ")

    if len(n) != ndim:
        raise ValueError("n must be a tuple with {ndim} elements")

    if len(b) != ndim:
        raise ValueError("b must be a tuple with {ndim} elements")

    axes = [jnp.linspace(-bi, bi, ni) for bi, ni in zip(b, n)]
    points = jnp.stack(jnp.meshgrid(*axes, indexing="ij"), axis=-1)
    points = points.reshape(-1, ndim)
    return Mesh(points, axes=axes)


def density(basis: Basis, mesh: Mesh, P: Optional[FloatNxN] = None) -> FloatN:
    P = jnp.diag(basis.occupancy) if P is None else P
    orbitals = basis(mesh.points)
    return jnp.einsum("ij,pi,pj->p", P, orbitals, orbitals)


def density_and_grad(
    basis: Basis, mesh: Mesh, P: Optional[FloatNxN] = None
) -> Tuple[FloatN, FloatNx3]:
    def f(points):
        return density(basis, eqx.combine(points, rest), P)

    points, rest = eqx.partition(mesh, lambda x: id(x) == id(mesh.points))
    rho, df = vjp(f, points)
    grad_rho = df(jnp.ones_like(rho))[0].points
    return rho, grad_rho


def molecular_orbitals(
    basis: Basis, mesh: Mesh, C: Optional[FloatNxN] = None
) -> FloatN:
    C = jnp.eye(basis.num_orbitals) if C is None else C
    orbitals = basis(mesh.points) @ C
    return orbitals


def xcmesh_from_pyscf(structure: Structure, level: int = 3) -> Mesh:
    grids = dft.gen_grid.Grids(to_pyscf(structure))
    grids.level = level
    grids.build()
    return Mesh(points=grids.coords, weights=grids.weights)



"""Container for a linear combination of Gaussian Primitives (aka contraction)."""

from functools import partial
from typing import Tuple

import equinox as eqx
import jax.numpy as jnp
from jax import tree, vmap

from mess.primitive import Primitive, eval_primitive
from mess.types import FloatN, FloatNx3


class Orbital(eqx.Module):
    primitives: Tuple[Primitive]
    coefficients: FloatN

    @property
    def num_primitives(self) -> int:
        return len(self.primitives)

    def __call__(self, pos: FloatNx3) -> FloatN:
        pos = jnp.atleast_2d(pos)
        assert pos.ndim == 2 and pos.shape[1] == 3, "pos must have shape [N,3]"

        @partial(vmap, in_axes=(0, 0, None))
        def eval_orbital(p: Primitive, coef: float, pos: FloatNx3):
            return coef * eval_primitive(p, pos)

        batch = tree.map(lambda *xs: jnp.stack(xs), *self.primitives)
        out = jnp.sum(eval_orbital(batch, self.coefficients, pos), axis=0)
        return out

    @staticmethod
    def from_bse(center, alphas, lmn, coefficients):
        coefficients = coefficients.reshape(-1)
        assert len(coefficients) == len(alphas), "Expecting same size vectors!"
        p = [Primitive(center=center, alpha=a, lmn=lmn) for a in alphas]
        return Orbital(primitives=p, coefficients=coefficients)


def batch_orbitals(orbitals: Tuple[Orbital]):
    primitives = [p for o in orbitals for p in o.primitives]
    primitives = tree.map(lambda *xs: jnp.stack(xs), *primitives)
    coefficients = jnp.concatenate([o.coefficients for o in orbitals])
    orbital_index = jnp.concatenate([
        i * jnp.ones(o.num_primitives, dtype=jnp.int32) for i, o in enumerate(orbitals)
    ])
    return primitives, coefficients, orbital_index



"""automatic differentiation of atomic orbital integrals"""

from functools import partial
from typing import Callable

import equinox as eqx
import jax.numpy as jnp
from jax import grad, jit, tree, vmap
from jax.ops import segment_sum

from mess.basis import Basis
from mess.integrals import _kinetic_primitives, _nuclear_primitives, _overlap_primitives
from mess.primitive import Primitive
from mess.types import Float3, Float3xNxN


def grad_integrate_primitives(a: Primitive, b: Primitive, operator: Callable) -> Float3:
    def f(center):
        return operator(eqx.combine(center, arest), b)

    acenter, arest = eqx.partition(a, lambda x: id(x) == id(a.center))
    return grad(f)(acenter).center


def grad_overlap_primitives(a: Primitive, b: Primitive) -> Float3:
    return grad_integrate_primitives(a, b, _overlap_primitives)


def grad_kinetic_primitives(a: Primitive, b: Primitive) -> Float3:
    return grad_integrate_primitives(a, b, _kinetic_primitives)


def grad_nuclear_primitives(a: Primitive, b: Primitive, c: Float3) -> Float3:
    def n(lhs, rhs):
        return _nuclear_primitives(lhs, rhs, c)

    return grad_integrate_primitives(a, b, n)


@partial(jit, static_argnums=1)
def grad_integrate_basis(basis: Basis, operator: Callable) -> Float3xNxN:
    def take_primitives(indices):
        p = tree.map(lambda x: jnp.take(x, indices, axis=0), basis.primitives)
        c = jnp.take(basis.coefficients, indices)
        return p, c

    ii, jj = jnp.meshgrid(*[jnp.arange(basis.num_primitives)] * 2, indexing="ij")
    lhs, cl = take_primitives(ii.reshape(-1))
    rhs, cr = take_primitives(jj.reshape(-1))
    out = vmap(operator)(lhs, rhs)

    out = cl * cr * out.T
    out = out.reshape(3, basis.num_primitives, basis.num_primitives)
    out = jnp.rollaxis(out, 1)
    out = segment_sum(out, basis.orbital_index, num_segments=basis.num_orbitals)
    out = jnp.rollaxis(out, -1)
    out = segment_sum(out, basis.orbital_index, num_segments=basis.num_orbitals)
    return jnp.rollaxis(out, -1)


def grad_overlap_basis(basis: Basis) -> Float3xNxN:
    return grad_integrate_basis(basis, grad_overlap_primitives)


def grad_kinetic_basis(basis: Basis) -> Float3xNxN:
    return grad_integrate_basis(basis, grad_kinetic_primitives)


def grad_nuclear_basis(basis: Basis) -> Float3xNxN:
    def n(atomic_number, position):
        def op(pi, pj):
            return atomic_number * grad_nuclear_primitives(pi, pj, position)

        return grad_integrate_basis(basis, op)

    out = vmap(n)(basis.structure.atomic_number, basis.structure.position)
    return jnp.sum(out, axis=0)



"""Core functions for common approximations to the exchange-correlation functional"""

import jax
import numpy as np
import jax.numpy as jnp
import jax.numpy.linalg as jnl
from mess.types import FloatN, FloatNx3


def fzeta(z):
    u = (1 + z) ** (4 / 3) + (1 - z) ** (4 / 3) - 2
    v = 2 * (2 ** (1 / 3) - 1)
    return u / v


def d2fdz20():
    f = jax.grad(jax.grad(fzeta))
    return float(f(0.0))


F2 = d2fdz20()

default_threshold = 1e-15


def lda_exchange(rho: FloatN, threshold: float = default_threshold) -> FloatN:
    mask = rho > threshold
    rho = jnp.where(mask, rho, 0.0)
    Cx = (3 / 4) * (3 / np.pi) ** (1 / 3)
    eps_x = -Cx * rho ** (1 / 3)
    eps_x = jnp.where(mask, eps_x, 0.0)
    return eps_x


def lda_correlation_vwn(
    rho: FloatN, threshold: float = default_threshold, use_rpa: bool = True
) -> FloatN:
    A, x0, b, c = vwn_coefs(use_rpa)

    # avoid divide by zero when rho = 0 by replacing with 1.0
    mask = jnp.abs(rho) > threshold
    rho = jnp.where(mask, rho, 1.0)
    rs = jnp.power(3 / (4 * jnp.pi * rho), 1 / 3).reshape(-1, 1)
    x = jnp.sqrt(rs).reshape(-1, 1)
    X = rs + b * x + c
    X0 = x0**2 + b * x0 + c
    Q = np.sqrt(4 * c - b**2)

    u = jnp.log(x**2 / X) + 2 * b / Q * jnp.arctan(Q / (2 * x + b))
    v = jnp.log((x - x0) ** 2 / X) + 2 * (b + 2 * x0) / Q * jnp.arctan(Q / (2 * x + b))
    ec = A * (u - b * x0 / X0 * v)
    e0, e1, alpha = ec.T
    beta = F2 * (e1 - e0) / alpha - 1
    z = jnp.zeros_like(rho)  # restricted ks, should be rho_up - rho_down
    eps_c = e0 + alpha * fzeta(z) / F2 * (1 + beta * z**4)
    eps_c = jnp.where(mask, eps_c, 0.0)
    return eps_c


def vwn_coefs(use_rpa: bool = True):
    # paramagnetic (eps_0) / ferromagnetic (eps_1) / spin stiffness (alpha)
    A = np.array([0.0310907, 0.5 * 0.0310907, -1 / (6 * np.pi**2)])

    if use_rpa:
        x0 = np.array([-0.409286, -0.743294, -0.228344])
        b = np.array([13.0720, 20.1231, 1.06835])
        c = np.array([42.7198, 101.578, 11.4813])
    else:
        # https://math.nist.gov/DFTdata/atomdata/node5.html
        x0 = np.array([-0.10498, -0.32500, -4.75840e-3])
        b = np.array([3.72744, 7.06042, 1.13107])
        c = np.array([12.9352, 18.0578, 13.0045])
    return A, x0, b, c


def lda_correlation_pw(rho: FloatN, threshold: float = default_threshold) -> FloatN:
    p = np.ones(3)
    A = np.array([0.031091, 0.015545, 0.016887])
    a1 = np.array([0.21370, 0.20548, 0.11125])
    b1 = np.array([7.5957, 14.1189, 10.357])
    b2 = np.array([3.5876, 6.1977, 3.6231])
    b3 = np.array([1.6382, 3.3662, 0.88026])
    b4 = np.array([0.49294, 0.62517, 0.49671])

    # avoid divide by zero when rho = 0 by replacing with 1.0
    mask = jnp.abs(rho) > threshold
    rho = jnp.where(mask, rho, 1.0)
    rs = jnp.power(3 / (4 * jnp.pi * rho), 1 / 3).reshape(-1, 1)
    v = 2 * A * (b1 * jnp.sqrt(rs) + b2 * rs + b3 * rs ** (3 / 2) + b4 * rs ** (p + 1))
    G = -2 * A * (1 + a1 * rs) * jnp.log(1 + 1 / v)
    e0, e1, alpha = G.T
    beta = F2 * (e1 - e0) / alpha - 1
    z = jnp.zeros_like(rho)  # restricted ks, should be rho_up - rho_down
    eps_c = e0 + alpha * fzeta(z) / F2 * (1 + beta * z**4)
    eps_c = jnp.where(mask, eps_c, 0.0)
    return eps_c


def gga_exchange_b88(
    rho: FloatN, grad_rho: FloatNx3, threshold: float = default_threshold
) -> FloatN:
    beta = jnp.asarray(0.0042 * 2 ** (1 / 3))
    # avoid divide by zero when rho = 0 by replacing with 1.0
    mask = jnp.abs(rho) > threshold
    rho = jnp.where(mask, rho, 1.0)
    x = jnl.norm(grad_rho, axis=1) / rho ** (4 / 3)
    d = 1 + 6 * beta * x * jnp.arcsinh(2 ** (1 / 3) * x)
    eps_x = lda_exchange(rho) - beta * rho ** (1 / 3) * x**2 / d
    eps_x = jnp.where(mask, eps_x, 0.0)
    return eps_x


def gga_exchange_pbe(
    rho: FloatN, grad_rho: FloatNx3, threshold: float = default_threshold
) -> FloatN:
    beta = np.asarray(0.066725)  # Eq 4
    mu = beta * np.pi**2 / 3  # Eq 12
    kappa = np.asarray(0.8040)  # Eq 14

    # avoid divide by zero when rho = 0 by replacing with 1.0
    mask = jnp.abs(rho) > threshold
    rho = jnp.where(mask, rho, 1.0)
    kf = (3 * np.pi**2 * rho) ** (1 / 3)
    s = jnl.norm(grad_rho, axis=1) / (2 * kf * rho)
    F = 1 + kappa - kappa / (1 + mu * s**2 / kappa)
    F = jnp.where(mask, F, 0.0)
    return lda_exchange(rho) * F


def gga_correlation_pbe(
    rho: FloatN, grad_rho: FloatNx3, threshold: float = default_threshold
) -> FloatN:
    beta = np.asarray(0.066725)
    gamma = (1 - np.log(2.0)) / np.pi**2
    z = jnp.zeros_like(rho)  # restricted ks, should be (rho_up - rho_down) / rho
    phi = 0.5 * (jnp.power(1 + z, 2 / 3) + jnp.power(1 - z, 2 / 3))
    ec_pw = lda_correlation_pw(rho, threshold)
    # avoid divide by zero when rho = 0 by replacing with 1.0
    mask = jnp.abs(rho) > threshold
    rho = jnp.where(mask, rho, 1.0)
    mask_ecpw = jnp.where(mask, ec_pw, 1.0)
    A = beta / gamma * (jnp.exp(-mask_ecpw / (gamma * phi**3)) - 1) ** -1  # Eq 8
    kf = (3 * np.pi**2 * rho) ** (1 / 3)
    ks = jnp.sqrt(4 * kf / np.pi)
    t = jnl.norm(grad_rho, axis=1) / (2 * phi * ks * rho)
    u = 1 + beta / gamma * t**2 * (1 + A * t**2) / (1 + A * t**2 + A**2 * t**4)
    H = gamma * phi**3 * jnp.log(u)  # Eq 7
    H = jnp.where(mask, H, 0.0)
    return ec_pw + H


def gga_correlation_lyp(
    rho: FloatN, grad_rho: FloatNx3, threshold: float = default_threshold
) -> FloatN:
    a = np.asarray(0.04918)
    b = np.asarray(0.132)
    c = np.asarray(0.2533)
    d = np.asarray(0.349)
    CF = 0.3 * (3 * np.pi**2) ** (2 / 3)

    # avoid divide by zero when rho = 0 by replacing with 1.0
    mask = jnp.abs(rho) > threshold
    rho = jnp.where(mask, rho, 1.0)
    v = 1 + d * rho ** (-1 / 3)
    omega = jnp.exp(-c * rho ** (-1 / 3)) / v * rho ** (-11 / 3)
    delta = c * rho ** (-1 / 3) + d * rho ** (-1 / 3) / v
    g = (1 / 24 + 7 * delta / 72) * rho * jnl.norm(grad_rho, axis=1) ** 2

    eps_c = -a / v - a * b * omega * (CF * rho ** (11 / 3) - g)
    eps_c = jnp.where(mask, eps_c, 0.0)
    return eps_c



import jax.numpy as jnp
import numpy as np
import pytest
from numpy.testing import assert_allclose

from mess.basis import basisset
from mess.interop import to_pyscf
from mess.mesh import density, density_and_grad, uniform_mesh
from mess.structure import molecule, nuclear_energy


@pytest.mark.parametrize("basis_name", ["sto-3g", "6-31g**"])
def test_to_pyscf(basis_name):
    mol = molecule("water")
    basis = basisset(mol, basis_name)
    pyscf_mol = to_pyscf(mol, basis_name)
    assert basis.num_orbitals == pyscf_mol.nao


def test_gto():
    from pyscf.dft.numint import eval_rho, eval_ao
    from jax.experimental import enable_x64

    with enable_x64(True):
        # Run these comparisons to PySCF in fp64
        # Atomic orbitals
        basis_name = "6-31+g"
        structure = molecule("water")
        basis = basisset(structure, basis_name)
        mesh = uniform_mesh()
        actual = basis(mesh.points)

        mol = to_pyscf(structure, basis_name)
        expect_ao = eval_ao(mol, np.asarray(mesh.points))
        assert_allclose(actual, expect_ao, atol=1e-7)

        # Density Matrix
        mf = mol.RKS()
        mf.kernel()
        C = jnp.array(mf.mo_coeff)
        P = basis.density_matrix(C)
        expect = jnp.array(mf.make_rdm1())
        assert_allclose(P, expect)

        # Electron density
        actual = density(basis, mesh, P)
        expect = eval_rho(mol, expect_ao, mf.make_rdm1(), xctype="lda")
        assert_allclose(actual, expect, atol=1e-7)

        # Electron density and gradient
        rho, grad_rho = density_and_grad(basis, mesh, P)
        ao_and_grad = eval_ao(mol, np.asarray(mesh.points), deriv=1)
        expect = eval_rho(mol, ao_and_grad, mf.make_rdm1(), xctype="gga")
        expect_rho = expect[0, :]
        expect_grad = expect[1:, :].T
        assert_allclose(rho, expect_rho, atol=1e-7)
        assert_allclose(grad_rho, expect_grad, atol=1e-6)


@pytest.mark.parametrize("name", ["water", "h2"])
def test_nuclear_energy(name):
    mol = molecule(name)
    actual = nuclear_energy(mol)
    expect = to_pyscf(mol).energy_nuc()
    assert_allclose(actual, expect)



"""Conversion between Bohr and Angstrom units

Note:
    MESS uses atomic units internally so these conversions are only necessary when
    working with external packages.
"""

from jaxtyping import Array

# Maximum value an individual component of the angular momentum lmn can take
# Used for static ahead-of-time compilation of functions involving lmn.
LMAX = 4

BOHR_PER_ANGSTROM = 1.0 / 0.529177210903


def to_angstrom(bohr_value: Array) -> Array:
    return bohr_value / BOHR_PER_ANGSTROM


def to_bohr(angstrom_value: Array) -> Array:
    return angstrom_value * BOHR_PER_ANGSTROM



"""(experimental) Gaussian orbital integrals without array padding."""

from functools import partial

import jax.numpy as jnp
from jax import jit

from mess.basis import Basis
from mess.integrals import integrate
from mess.primitive import Primitive, product
from mess.types import FloatNxN


@partial(jit, static_argnums=0)
def overlap_basis_zeropad(basis: Basis) -> FloatNxN:
    def op(a, b):
        return _overlap_primitives_zeropad(a, b, basis.max_L)

    return integrate(basis, op)


@partial(jit, static_argnums=2)
def overlap_context(i: int, j: int, max_L: int):
    from mess.special import binom, factorial2

    def gen():
        for s in range(max_L + 1):
            for t in range(2 * s + 1):
                yield s, t

    s, t = jnp.asarray(list(gen())).T
    mask = (2 * s - i <= t) & (t <= j)
    s = jnp.where(mask, s, 0)
    t = jnp.where(mask, t, 0)
    w = binom(i, 2 * s - t) * binom(j, t) * factorial2(2 * s - 1)
    return s, t, w


def _overlap_primitives_zeropad(a: Primitive, b: Primitive, max_L: int) -> float:
    def overlap_axis(i: int, j: int, a: float, b: float) -> float:
        s, t, w = overlap_context(i, j, max_L)
        out = w * a ** (i - (2 * s - t)) * b ** (j - t) / (2 * p.alpha) ** s
        return jnp.sum(out)

    p = product(a, b)
    pa = p.center - a.center
    pb = p.center - b.center
    out = jnp.power(jnp.pi / p.alpha, 1.5) * p.norm

    for ax in range(3):
        out *= overlap_axis(a.lmn[ax], b.lmn[ax], pa[ax], pb[ax])

    return out



"""Interoperation tools for working across MESS, PySCF."""

from typing import Tuple

import numpy as np
from periodictable import elements
from pyscf import gto

from mess.basis import Basis, basisset
from mess.structure import Structure
from mess.units import to_bohr
from mess.package_utils import requires_package


def to_pyscf(structure: Structure, basis_name: str = "sto-3g") -> "gto.Mole":
    mol = gto.Mole(unit="Bohr", spin=structure.num_electrons % 2, cart=True)
    mol.atom = [
        (symbol, pos)
        for symbol, pos in zip(structure.atomic_symbol, structure.position)
    ]
    mol.basis = basis_name
    mol.build(unit="Bohr")
    return mol


def from_pyscf(mol: "gto.Mole") -> Tuple[Structure, Basis]:
    atoms = [(elements.symbol(sym).number, pos) for sym, pos in mol.atom]
    atomic_number, position = [np.array(x) for x in zip(*atoms)]

    if mol.unit == "Angstrom":
        position = to_bohr(position)

    structure = Structure(atomic_number, position)

    basis = basisset(structure, basis_name=mol.basis)

    return structure, basis


@requires_package("pyquante2")
def from_pyquante(name: str) -> Structure:
    """Load molecular structure from pyquante2.geo.samples module

    Args:
        name (str): Possible names include ch4, c6h6, aspirin, caffeine, hmx, petn,
                    prozan, rdx, taxol, tylenol, viagara, zoloft

    Returns:
        Structure
    """
    from pyquante2.geo import samples

    pqmol = getattr(samples, name)
    atomic_number, position = zip(*[(a.Z, a.r) for a in pqmol])
    atomic_number, position = [np.asarray(x) for x in (atomic_number, position)]
    return Structure(atomic_number, position)



"""Special mathematical functions not readily available in JAX."""

from functools import partial
from itertools import combinations_with_replacement

import jax.numpy as jnp
import numpy as np
from jax import lax, vmap
from jax.ops import segment_sum
from jax.scipy.special import betaln, gammainc, gammaln, erfc

from mess.types import FloatN, IntN
from mess.units import LMAX


def factorial_fori(n: IntN, nmax: int = LMAX) -> IntN:
    def body_fun(i, val):
        return val * jnp.where(i <= n, i, 1)

    return lax.fori_loop(1, nmax + 1, body_fun, jnp.ones_like(n))


def factorial_gamma(n: IntN) -> IntN:
    """Appoximate factorial by evaluating the gamma function in log-space.

    This approximation is exact for small integers (n < 10).
    """
    approx = jnp.exp(gammaln(n + 1))
    return jnp.rint(approx)


def factorial_lookup(n: IntN, nmax: int = LMAX) -> IntN:
    N = np.cumprod(np.arange(1, nmax + 1))
    N = np.insert(N, 0, 1)
    N = jnp.array(N, dtype=jnp.uint32)
    return N.at[n.astype(jnp.uint32)].get()


factorial = factorial_gamma


def factorial2_fori(n: IntN, nmax: int = 2 * LMAX) -> IntN:
    def body_fun(i, val):
        return val * jnp.where((i <= n) & (n % 2 == i % 2), i, 1)

    return lax.fori_loop(1, nmax + 1, body_fun, jnp.ones_like(n))


def factorial2_lookup(n: IntN, nmax: int = 2 * LMAX) -> IntN:
    stop = nmax + 1 if nmax % 2 == 0 else nmax + 2
    N = np.arange(1, stop).reshape(-1, 2)
    N = np.cumprod(N, axis=0).reshape(-1)
    N = np.insert(N, 0, 1)
    N = jnp.array(N)
    n = jnp.maximum(n, 0)
    return N.at[n].get()


factorial2 = factorial2_lookup


def binom_beta(x: IntN, y: IntN) -> IntN:
    approx = 1.0 / ((x + 1) * jnp.exp(betaln(x - y + 1, y + 1)))
    return jnp.rint(approx)


def binom_fori(x: IntN, y: IntN, nmax: int = LMAX) -> IntN:
    bang = partial(factorial_fori, nmax=nmax)
    c = x * bang(x - 1) / (bang(y) * bang(x - y))
    return jnp.where(x == y, 1, c)


def binom_lookup(x: IntN, y: IntN, nmax: int = LMAX) -> IntN:
    bang = partial(factorial_lookup, nmax=nmax)
    c = x * bang(x - 1) / (bang(y) * bang(x - y))
    return jnp.where(x == y, 1, c)


binom = binom_lookup


def gammanu_gamma(nu: IntN, t: FloatN, epsilon: float = 1e-10) -> FloatN:
    """Eq 2.11 from THO but simplified using SymPy and converted to jax

        t, u = symbols("t u", real=True, positive=True)
        nu = Symbol("nu", integer=True, nonnegative=True)

        expr = simplify(integrate(u ** (2 * nu) * exp(-t * u**2), (u, 0, 1)))
        f = lambdify((nu, t), expr, modules="scipy")
        ?f

    We evaulate this in log-space to avoid overflow/nan
    """
    t = jnp.maximum(t, epsilon)
    x = nu + 0.5
    gn = jnp.log(0.5) - x * jnp.log(t) + jnp.log(gammainc(x, t)) + gammaln(x)
    return jnp.exp(gn)


def gammanu_series(nu: IntN, t: FloatN, num_terms: int = 128) -> FloatN:
    """Eq 2.11 from THO but simplified as derived in equation 19 of gammanu.ipynb"""
    an = nu + 0.5
    tn = 1 / an
    total = 1 / an

    for _ in range(num_terms):
        an = an + 1
        tn = tn * t / an
        total = total + tn

    return 0.5 * jnp.exp(-t) * total


def gammanu_frac_vmap(nu: IntN, t: FloatN, num_terms: int = 128) -> FloatN:
    def scalar_fn(nu):
        n = jnp.arange(1, num_terms + 1, dtype=t.dtype)
        terms = jnp.where(nu >= n, t**n / jnp.cumprod(n - 0.5), 0.0)
        q = erfc(jnp.sqrt(t)) + jnp.exp(-t) / jnp.sqrt(jnp.pi * t) * jnp.sum(terms)
        lnout = (
            jnp.log(0.5) - (nu + 0.5) * jnp.log(t) + jnp.log(1 - q) + gammaln(nu + 0.5)
        )
        return jnp.exp(lnout)

    out_flt = vmap(scalar_fn)(nu.reshape(-1))
    return out_flt.reshape(nu.shape)


def gammanu_lax_series(nu: IntN, t: FloatN) -> FloatN:
    def cond_fn(vals):
        return jnp.any(vals[0])

    def body_fn(vals):
        enabled, an, tn, total = vals
        an = an + 1
        tn = tn * (t / an)
        total = total + tn
        enabled = enabled & ((tn / total) > jnp.finfo(t.dtype).eps)
        return (enabled, an, tn, total)

    a0 = nu + 0.5
    t0 = 1.0 / a0
    total = 1.0 / t0

    init_vals = (
        jnp.ones(nu.shape, dtype=bool),
        a0,
        t0,
        total,
    )

    _, _, _, total = lax.while_loop(cond_fn, body_fn, init_vals)
    return jnp.exp(jnp.log(0.5) + jnp.log(total) - t)


def gammanu_lax_frac(nu: IntN, t: FloatN) -> FloatN:
    def cond_fn(vals):
        return jnp.any(vals[0])

    def body_fn(vals):
        enabled, term, n, q = vals
        term = term * t / (n - 0.5)
        enabled = nu >= n
        q = jnp.where(enabled, q + term, q)
        n = n + 1.0
        return (enabled, term, n, q)

    enabled = jnp.ones(nu.shape, dtype=bool)
    term = jnp.full(nu.shape, jnp.exp(-t) / jnp.sqrt(jnp.pi * t))
    n = jnp.ones(nu.shape)
    q = jnp.full(nu.shape, erfc(jnp.sqrt(t)))
    init_vals = (enabled, term, n, q)
    _, _, _, q = lax.while_loop(cond_fn, body_fn, init_vals)
    lnout = jnp.log(0.5) - (nu + 0.5) * jnp.log(t) + jnp.log(1 - q) + gammaln(nu + 0.5)
    return jnp.exp(lnout)


def gammanu_select(nu: IntN, t: FloatN, threshold: float = 50.0) -> FloatN:
    """Select between different implementation strategies for evaluation of gammanu

    Args:
        nu (IntN):
        t (FloatN):

    Returns:
        FloatN
    """
    y0 = 1 / (2 * nu + 1)
    y1 = gammanu_series(nu, jnp.minimum(t, threshold))
    y2 = (
        factorial2(2 * nu - 1)
        / 2 ** (nu + 1)
        * jnp.sqrt(jnp.pi / jnp.where(t >= threshold, t, 1.0) ** (2 * nu + 1))
    )

    return jnp.select(
        (t == 0, t < threshold, t >= threshold),
        (y0, y1, y2),
    )


gammanu = gammanu_select


def triu_indices(n: int):
    out = []
    for i in range(n):
        for j in range(i, n):
            out.append((i, j))
    i, j = np.asarray(out).T
    return i, j


def tril_indices(n: int):
    out = []
    for i in range(n):
        for j in range(i + 1):
            out.append((i, j))
    i, j = np.asarray(out).T
    return i, j


def allpairs_indices(n: int):
    pairs_gen = combinations_with_replacement(range(n), r=2)
    i, j = np.asarray(list(pairs_gen)).T
    return i, j


def binom_factor(i: int, j: int, a: float, b: float, lmax: int = LMAX) -> FloatN:
    """Eq. 15 from Augspurger JD, Dykstra CE. General quantum mechanical operators. An
    open-ended approach for one-electron integrals with Gaussian bases. Journal of
    computational chemistry. 1990 Jan;11(1):105-11.
    <https://doi.org/10.1002/jcc.540110113>
    """
    s, t = tril_indices(lmax + 1)
    apow = jnp.maximum(i - (s - t), 0)
    bpow = jnp.maximum(j - t, 0)
    out = binom(i, s - t) * binom(j, t) * a**apow * b**bpow
    mask = ((s - i) <= t) & (t <= j)
    out = jnp.where(mask, out, 0.0)
    return segment_sum(out, s, num_segments=lmax + 1)



from functools import partial

import numpy as np
import pytest
from jax.experimental import enable_x64
from numpy.testing import assert_allclose
from pyscf import dft

from mess.basis import basisset
from mess.mesh import density_and_grad, xcmesh_from_pyscf
from mess.structure import Structure
from mess.xcfunctional import (
    gga_correlation_lyp,
    gga_correlation_pbe,
    gga_exchange_b88,
    gga_exchange_pbe,
    lda_correlation_pw,
    lda_correlation_vwn,
    lda_exchange,
)

# Define test cases with a mapping from test identifier to test arguments
lda_cases = {
    "lda_exchange": (lda_exchange, "slater,"),
    "lda_correlation_vwn5": (partial(lda_correlation_vwn, use_rpa=False), ",vwn5"),
    "lda_correlation_vwn_rpa": (partial(lda_correlation_vwn, use_rpa=True), ",vwn_rpa"),
    "lda_correlation_pw": (lda_correlation_pw, ",lda_c_pw"),
}

gga_cases = {
    "gga_exchange_b88": (gga_exchange_b88, "gga_x_b88,"),
    "gga_exchange_pbe": (gga_exchange_pbe, "gga_x_pbe,"),
    "gga_correlation_pbe": (gga_correlation_pbe, ",gga_c_pbe"),
    "gga_correlation_lyp": (gga_correlation_lyp, ",gga_c_lyp"),
}


@pytest.fixture
def helium_density():
    with enable_x64(True):
        mol = Structure(np.asarray(2), np.zeros(3))
        basis_name = "6-31g"
        basis = basisset(mol, basis_name)
        mesh = xcmesh_from_pyscf(mol)
        rho, grad_rho = [np.asarray(t) for t in density_and_grad(basis, mesh)]
        yield rho, grad_rho


@pytest.mark.parametrize("xcfunc,scfstr", lda_cases.values(), ids=lda_cases.keys())
def test_lda(helium_density, xcfunc, scfstr):
    rho, _ = helium_density
    actual = xcfunc(rho)
    expect = dft.libxc.eval_xc(scfstr, rho)[0]
    assert_allclose(actual, expect, atol=1e-7)


@pytest.mark.parametrize("xcfunc,scfstr", gga_cases.values(), ids=gga_cases.keys())
def test_gga(helium_density, xcfunc, scfstr):
    rho, grad_rho = helium_density
    scfin = np.concatenate([rho[:, None], grad_rho], axis=1).T

    actual = xcfunc(rho, grad_rho)
    expect = dft.libxc.eval_xc(scfstr, scfin, deriv=1)[0]
    assert_allclose(actual, expect, atol=1e-6)



import importlib
from functools import wraps
from typing import Any, Callable, TypeVar

F = TypeVar("F", bound=Callable[..., Any])


class MissingOptionalDependencyError(BaseException):
    """
    An exception raised when an optional dependency is required
    but cannot be found.

    Attributes
    ----------
    library_name
        The name of the missing library.
    """

    def __init__(self, library_name: str):
        """

        Parameters
        ----------
        library_name
            The name of the missing library.
        license_issue
            Whether the library was importable but was unusable due
            to a missing license.
        """

        message = f"The required {library_name} module could not be imported."

        super(MissingOptionalDependencyError, self).__init__(message)

        self.library_name = library_name


def has_package(package_name: str) -> bool:
    """
    Helper function to generically check if a Python package is installed.
    Intended to be used to check for optional dependencies.

    Parameters
    ----------
    package_name : str
        The name of the Python package to check the availability of

    Returns
    -------
    package_available : bool
        Boolean indicator if the package is available or not

    Examples
    --------
    >>> has_numpy = has_package('numpy')
    >>> has_numpy
    True
    >>> has_foo = has_package('other_non_installed_package')
    >>> has_foo
    False
    """
    try:
        importlib.import_module(package_name)
    except ModuleNotFoundError:
        return False
    return True


def requires_package(package_name: str) -> Callable[..., Any]:
    """
    Helper function to denote that a funciton requires some optional
    dependency. A function decorated with this decorator will raise
    `MissingOptionalDependencyError` if the package is not found by
    `importlib.import_module()`.

    Parameters
    ----------
    package_name : str
        The name of the module to be imported.

    Raises
    ------
    MissingOptionalDependencyError

    """

    def inner_decorator(function: F) -> F:
        @wraps(function)
        def wrapper(*args, **kwargs):
            import importlib

            try:
                importlib.import_module(package_name)
            except ImportError:
                raise MissingOptionalDependencyError(library_name=package_name)
            except Exception as e:
                raise e

            return function(*args, **kwargs)

        return wrapper

    return inner_decorator



import jax.numpy as jnp
import numpy as np
import pytest
from numpy.testing import assert_allclose

from mess.basis import basisset
from mess.integrals import (
    eri_basis,
    eri_basis_sparse,
    eri_primitives,
    kinetic_basis,
    kinetic_primitives,
    nuclear_basis,
    nuclear_primitives,
    overlap_basis,
    overlap_primitives,
)
from mess.interop import to_pyscf
from mess.primitive import Primitive
from mess.structure import molecule
from conftest import is_mem_limited


def test_overlap():
    # Exercise 3.21 of "Modern quantum chemistry: introduction to advanced
    # electronic structure theory."" by Szabo and Ostlund
    alpha = 0.270950 * 1.24 * 1.24
    a = Primitive(alpha=alpha)
    b = Primitive(alpha=alpha, center=jnp.array([1.4, 0.0, 0.0]))
    assert_allclose(overlap_primitives(a, a), 1.0, atol=1e-5)
    assert_allclose(overlap_primitives(b, b), 1.0, atol=1e-5)
    assert_allclose(overlap_primitives(b, a), 0.6648, atol=1e-5)


@pytest.mark.parametrize("basis_name", ["sto-3g", "6-31+g", "6-31+g*"])
def test_water_overlap(basis_name):
    basis = basisset(molecule("water"), basis_name)
    actual_overlap = overlap_basis(basis)

    # Note: PySCF doesn't appear to normalise d basis functions in cartesian basis
    scfmol = to_pyscf(molecule("water"), basis_name=basis_name)
    expect_overlap = scfmol.intor("int1e_ovlp_cart")
    n = 1 / np.sqrt(np.diagonal(expect_overlap))
    expect_overlap = n[:, None] * n[None, :] * expect_overlap
    assert_allclose(actual_overlap, expect_overlap, atol=1e-5)


def test_kinetic():
    # PyQuante test case for kinetic primitive integral
    p = Primitive()
    assert_allclose(kinetic_primitives(p, p), 1.5, atol=1e-5)

    # Reproduce the kinetic energy matrix for H2 using STO-3G basis set
    # See equation 3.230 of "Modern quantum chemistry: introduction to advanced
    # electronic structure theory."" by Szabo and Ostlund
    h2 = molecule("h2")
    basis = basisset(h2, "sto-3g")
    actual = kinetic_basis(basis)
    expect = np.array([[0.7600, 0.2365], [0.2365, 0.7600]])
    assert_allclose(actual, expect, atol=1e-4)


@pytest.mark.parametrize(
    "basis_name",
    [
        "sto-3g",
        "6-31+g",
        pytest.param(
            "6-31+g*", marks=pytest.mark.xfail(reason="Cartesian norm problem?")
        ),
    ],
)
def test_water_kinetic(basis_name):
    basis = basisset(molecule("water"), basis_name)
    actual = kinetic_basis(basis)

    expect = to_pyscf(molecule("water"), basis_name=basis_name).intor("int1e_kin_cart")
    assert_allclose(actual, expect, atol=1e-4)


def test_nuclear():
    # PyQuante test case for nuclear attraction integral
    p = Primitive()
    c = jnp.zeros(3)
    assert_allclose(nuclear_primitives(p, p, c), -1.595769, atol=1e-5)

    # Reproduce the nuclear attraction matrix for H2 using STO-3G basis set
    # See equation 3.231 and 3.232 of Szabo and Ostlund
    h2 = molecule("h2")
    basis = basisset(h2, "sto-3g")
    actual = nuclear_basis(basis)
    expect = np.array([
        [[-1.2266, -0.5974], [-0.5974, -0.6538]],
        [[-0.6538, -0.5974], [-0.5974, -1.2266]],
    ])

    assert_allclose(actual, expect, atol=1e-4)


def test_water_nuclear():
    basis_name = "sto-3g"
    h2o = molecule("water")
    basis = basisset(h2o, basis_name)
    actual = nuclear_basis(basis).sum(axis=0)
    expect = to_pyscf(h2o, basis_name=basis_name).intor("int1e_nuc_cart")
    assert_allclose(actual, expect, atol=1e-3)


def test_eri():
    # PyQuante test cases for ERI
    a, b, c, d = [Primitive()] * 4
    assert_allclose(eri_primitives(a, b, c, d), 1.128379, atol=1e-5)

    c, d = [Primitive(lmn=jnp.array([1, 0, 0]))] * 2
    assert_allclose(eri_primitives(a, b, c, d), 0.940316, atol=1e-5)

    # H2 molecule in sto-3g: See equation 3.235 of Szabo and Ostlund
    h2 = molecule("h2")
    basis = basisset(h2, "sto-3g")

    actual = eri_basis(basis)
    expect = np.empty((2, 2, 2, 2), dtype=np.float32)
    expect[0, 0, 0, 0] = expect[1, 1, 1, 1] = 0.7746
    expect[0, 0, 1, 1] = expect[1, 1, 0, 0] = 0.5697
    expect[1, 0, 0, 0] = expect[0, 0, 0, 1] = 0.4441
    expect[0, 1, 0, 0] = expect[0, 0, 1, 0] = 0.4441
    expect[0, 1, 1, 1] = expect[1, 1, 1, 0] = 0.4441
    expect[1, 0, 1, 1] = expect[1, 1, 0, 1] = 0.4441
    expect[1, 0, 1, 0] = expect[0, 1, 1, 0] = 0.2970
    expect[0, 1, 0, 1] = expect[1, 0, 0, 1] = 0.2970
    assert_allclose(actual, expect, atol=1e-4)


@pytest.mark.parametrize("sparse", [True, False])
@pytest.mark.skipif(is_mem_limited(), reason="Not enough host memory!")
def test_water_eri(sparse):
    basis_name = "sto-3g"
    h2o = molecule("water")
    basis = basisset(h2o, basis_name)
    actual = eri_basis_sparse(basis) if sparse else eri_basis(basis)
    aosym = "s8" if sparse else "s1"
    expect = to_pyscf(h2o, basis_name=basis_name).intor("int2e_cart", aosym=aosym)
    assert_allclose(actual, expect, atol=1e-4)



"""Types used throughout MESS

Note:
    ``N`` represents the number of atomic orbitals.
"""

from functools import partial
from typing import Tuple, Callable

import jax.numpy as jnp
from jax import config
from jaxtyping import Array, Float, Int

Float3 = Float[Array, "3"]
FloatNx3 = Float[Array, "N 3"]
FloatN = Float[Array, "N"]
Float3xNxN = Float[Array, "3 N N"]
FloatNxN = Float[Array, "N N"]
FloatNxM = Float[Array, "N M"]
Int3 = Int[Array, "3"]
IntN = Int[Array, "N"]

MeshAxes = Tuple[FloatN, FloatN, FloatN]

asintarray = partial(jnp.asarray, dtype=jnp.int32)

OrthNormTransform = Callable[[FloatNxN], FloatNxN]


def default_fptype():
    return jnp.float64 if config.x64_enabled else jnp.float32


asfparray = partial(jnp.asarray, dtype=default_fptype())



"""Function decorators to automate converting between numeric formats."""

from functools import wraps
from typing import Callable

import jax.numpy as jnp
import numpy as np
from jax.experimental import enable_x64
from jaxtyping import Array


def apply_fpcast(v: Array, dtype: np.dtype):
    if isinstance(v, jnp.ndarray) and np.issubdtype(v, np.floating):
        return v.astype(dtype)

    return v


def fpcast(func: Callable, dtype=jnp.float32):
    @wraps(func)
    def wrapper(*args, **kwargs):
        inputs = [apply_fpcast(v, dtype) for v in args]
        outputs = func(*inputs, **kwargs)
        return outputs

    return wrapper


def compare_fp32_to_fp64(func: Callable):
    @wraps(func)
    def wrapper(*args, **kwargs):
        with enable_x64():
            outputs_fp32 = fpcast(func, dtype=jnp.float32)(*args, **kwargs)
            outputs_fp64 = fpcast(func, dtype=jnp.float64)(*args, **kwargs)
            print_compare(func.__name__, outputs_fp32, outputs_fp64)
            return outputs_fp32

    return wrapper


def print_compare(name: str, fp32, fp64):
    fp32 = [fp32] if isinstance(fp32, jnp.ndarray) else fp32
    fp64 = [fp64] if isinstance(fp64, jnp.ndarray) else fp64

    for idx, (low, high) in enumerate(zip(fp32, fp64)):
        low = np.asarray(low).astype(np.float64)
        high = np.asarray(high)
        print(f"{name} output {idx} has max |fp64 - fp32| = {np.abs(high - low).max()}")



"""Visualisations of molecular structures and volumetric data."""

import numpy as np
import py3Dmol
from more_itertools import chunked
from numpy.typing import NDArray

from mess.structure import Structure
from mess.types import MeshAxes
from mess.units import to_angstrom


def plot_molecule(view: py3Dmol.view, structure: Structure) -> py3Dmol.view:
    """Plots molecular structure.

    Args:
        view (py3Dmol.view): py3DMol View to which to add visualizer
        structure (Structure): molecular structure

    Returns:
        py3DMol View object

    """
    xyz = f"{structure.num_atoms}\n\n"
    sym = structure.atomic_symbol
    pos = to_angstrom(structure.position)

    for i in range(structure.num_atoms):
        r = np.array2string(pos[i, :], separator="\t")[1:-1]
        xyz += f"{sym[i]}\t{r}\n"

    view.addModel(xyz)
    style = "stick" if structure.num_atoms > 1 else "sphere"
    view.setStyle({style: {"radius": 0.1}})
    return view


def plot_volume(view: py3Dmol.view, value: NDArray, axes: MeshAxes):
    """Plots volumetric data value with molecular structure.

    Volumetric render using https://3dmol.csb.pitt.edu/doc/VolumetricRendererSpec.html

    Args:
        view (py3Dmol.view): py3DMol View to which to add visualizer
        value (NDArray): the volume data to render
        axes (MeshAxes): the axes over which the data was sampled.

    Returns:
        py3DMol View object

    """

    s = cube_data(value, axes)
    view.addVolumetricData(s, "cube", build_transferfn(value))
    return view


def plot_isosurfaces(
    view: py3Dmol.view, value: NDArray, axes: MeshAxes, percentiles=[95, 75]
):
    """Plots volumetric data value with molecular structure.

    IsoSurface render using https://3dmol.csb.pitt.edu/doc/IsoSurfaceSpec.html

    Args:
        view (py3Dmol.view): py3DMol View to which to add visualizer
        value (NDArray): the volume data to render
        axes (MeshAxes): the axes over which the data was sampled.
        percentiles (tuple): percentiles at which to draw isosurfaces

    Returns:
        py3DMol View object

    Note:
        3Dmol does not currently implement full transparency, so only two
        percentiles are accepted, with the inner one being rendered with full opacity.
        - https://github.com/3dmol/3Dmol.js/issues/224
    """
    assert len(percentiles) == 2

    voldata_str = cube_data(value, axes)

    v = np.percentile(np.abs(value), tuple(reversed(sorted(percentiles))))
    for sign in [-1, 1]:
        for isovalind in (0, 1):
            isoval = sign * v[isovalind]
            tf = {
                "isoval": isoval,
                "smoothness": 2,
                "opacity": 0.9 if isovalind == 1 else 1.0,
                "voldata": voldata_str,
                "volformat": "cube",
                "volscheme": {"gradient": "rwb", "min": -v[0], "max": v[0]},
            }
            view.addVolumetricData(voldata_str, "cube", tf)

    return view


def cube_data(value: NDArray, axes: MeshAxes) -> str:
    """Generate the cube file format as a string.  See:

      https://paulbourke.net/dataformats/cube/

    Args:
        value (NDArray): the volume data to serialise in the cube format
        axes (MeshAxes): the axes over which the data was sampled

    Returns:
        str: cube format representation of the volumetric data.
    """
    # The first two lines of the header are comments, they are generally ignored by
    # parsing packages or used as two default labels.
    fmt = "cube format\n\n"

    axes = [to_angstrom(ax) for ax in axes]
    x, y, z = axes
    # The third line has the number of atoms included in the file followed by the
    # position of the origin of the volumetric data.
    fmt += f"0 {cube_format_vec([x[0], y[0], z[0]])}\n"

    # The next three lines give the number of voxels along each axis (x, y, z)
    # followed by the axis vector. Note this means the volume need not be aligned
    # with the coordinate axis, indeed it also means it may be sheared although most
    # volumetric packages won't support that.
    # The length of each vector is the length of the side of the voxel thus allowing
    # non cubic volumes.
    # If the sign of the number of voxels in a dimension is positive then the
    # units are Bohr, if negative then Angstroms.
    nx, ny, nz = [ax.shape[0] for ax in axes]
    dx, dy, dz = [ax[1] - ax[0] for ax in axes]
    fmt += f"{nx} {cube_format_vec([dx, 0.0, 0.0])}\n"
    fmt += f"{ny} {cube_format_vec([0.0, dy, 0.0])}\n"
    fmt += f"{nz} {cube_format_vec([0.0, 0.0, dz])}\n"

    # The last section in the header is one line for each atom consisting of 5
    # numbers, the first is the atom number, the second is the charge, and the last
    # three are the x,y,z coordinates of the atom center.
    pass  # Number of atoms = 0 above

    # The volumetric data is straightforward, one floating point number for each
    # volumetric element.  Traditionally the grid is arranged with the x axis as
    # the outer loop and the z axis as the inner loop
    for vals in chunked(value, 6):
        fmt += f"{cube_format_vec(vals)}\n"

    return fmt


def cube_format_vec(vals):
    """
    From https://paulbourke.net/dataformats/cube, floats are formatted 12.6
    """
    # Benchmarks showed this is 4x faster than numpy.printoptions...
    return " ".join([f"{v:12.6f}" for v in vals])


def build_transferfn(value: NDArray) -> dict:
    """Generate the 3dmol.js transferfn argument for a particular value.

    Tries to set isovalues to capture main features of the volume data.

    Args:
        value (NDArray): the volume data.

    Returns:
        dict: containing transferfn
    """
    v = np.percentile(value, [99.9, 75])
    a = [0.02, 0.0005]
    return {
        "transferfn": [
            {"color": "blue", "opacity": a[0], "value": -v[0]},
            {"color": "blue", "opacity": a[1], "value": -v[1]},
            {"color": "white", "opacity": 0.0, "value": 0.0},
            {"color": "red", "opacity": a[1], "value": v[1]},
            {"color": "red", "opacity": a[0], "value": v[0]},
        ]
    }



"""basis sets of Gaussian type orbitals"""

from typing import Tuple

import equinox as eqx
import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax import jit
from jax.ops import segment_sum
from tabulate import tabulate

from mess.orbital import Orbital, batch_orbitals
from mess.primitive import Primitive
from mess.structure import Structure
from mess.types import FloatN, FloatNx3, FloatNxM, FloatNxN, IntN, default_fptype


class Basis(eqx.Module):
    orbitals: Tuple[Orbital]
    structure: Structure
    primitives: Primitive
    coefficients: FloatN
    orbital_index: IntN
    basis_name: str = eqx.field(static=True)
    max_L: int = eqx.field(static=True)

    @property
    def num_orbitals(self) -> int:
        return len(self.orbitals)

    @property
    def num_primitives(self) -> int:
        return sum(ao.num_primitives for ao in self.orbitals)

    @property
    def occupancy(self) -> FloatN:
        # Assumes uncharged systems in restricted Kohn-Sham
        occ = jnp.full(self.num_orbitals, 2.0)
        mask = occ.cumsum() > self.structure.num_electrons
        occ = jnp.where(mask, 0.0, occ)
        return occ

    def to_dataframe(self) -> pd.DataFrame:
        def fixer(x):
            # simple workaround for storing 2d array as a pandas column
            return [x[i, :] for i in range(x.shape[0])]

        df = pd.DataFrame()
        df["orbital"] = self.orbital_index
        df["coefficient"] = self.coefficients
        df["norm"] = self.primitives.norm
        df["center"] = fixer(self.primitives.center)
        df["lmn"] = fixer(self.primitives.lmn)
        df["alpha"] = self.primitives.alpha
        df.index.name = "primitive"
        return df

    def density_matrix(self, C: FloatNxN) -> FloatNxN:
        """Evaluate the density matrix from the molecular orbital coefficients

        Args:
            C (FloatNxN): the molecular orbital coefficients

        Returns:
            FloatNxN: the density matrix.
        """
        return jnp.einsum("k,ik,jk->ij", self.occupancy, C, C)

    @jit
    def __call__(self, pos: FloatNx3) -> FloatNxM:
        prim = self.coefficients[jnp.newaxis, :] * self.primitives(pos)
        orb = segment_sum(prim.T, self.orbital_index, num_segments=self.num_orbitals)
        return orb.T

    def __repr__(self) -> str:
        df = self.to_dataframe()
        headers = ["primitive"] + df.columns.to_list()
        return tabulate(df, headers)

    def _repr_html_(self) -> str | None:
        df = self.to_dataframe()
        return df._repr_html_()

    def __hash__(self) -> int:
        return hash(self.primitives)


def basisset(structure: Structure, basis_name: str = "sto-3g") -> Basis:
    """Factory function for building a basis set for a structure.

    Args:
        structure (Structure): Used to define the basis function parameters.
        basis_name (str, optional): Basis set name to look up on the
            `basis set exchange <https://www.basissetexchange.org/>`_.
            Defaults to ``sto-3g``.

    Returns:
        Basis constructed from inputs
    """
    from basis_set_exchange import get_basis
    from basis_set_exchange.sort import sort_basis

    # fmt: off
    LMN_MAP = {
        0: [(0, 0, 0)],
        1: [(1, 0, 0), (0, 1, 0), (0, 0, 1)],
        2: [(2, 0, 0), (1, 1, 0), (1, 0, 1), (0, 2, 0), (0, 1, 1), (0, 0, 2)],
        3: [(3, 0, 0), (2, 1, 0), (2, 0, 1), (1, 2, 0), (1, 1, 1),
            (1, 0, 2), (0, 3, 0), (0, 2, 1), (0, 1, 2), (0, 0, 3)],
    }
    # fmt: on

    bse_basis = get_basis(
        basis_name,
        elements=structure.atomic_number.tolist(),
        uncontract_spdf=True,
        uncontract_general=True,
    )
    bse_basis = sort_basis(bse_basis)["elements"]
    orbitals = []

    for a in range(structure.num_atoms):
        center = structure.position[a, :]
        shells = bse_basis[str(structure.atomic_number[a])]["electron_shells"]

        for s in shells:
            for lmn in LMN_MAP[s["angular_momentum"][0]]:
                ao = Orbital.from_bse(
                    center=center,
                    alphas=np.array(s["exponents"], dtype=default_fptype()),
                    lmn=np.array(lmn, dtype=np.int32),
                    coefficients=np.array(s["coefficients"], dtype=default_fptype()),
                )
                orbitals.append(ao)

    primitives, coefficients, orbital_index = batch_orbitals(orbitals)

    return Basis(
        orbitals=orbitals,
        structure=structure,
        primitives=primitives,
        coefficients=coefficients,
        orbital_index=orbital_index,
        basis_name=basis_name,
        max_L=int(np.max(primitives.lmn)),
    )


def basis_iter(basis: Basis):
    from jax import tree

    from mess.special import triu_indices

    def take_primitives(indices):
        p = tree.map(lambda x: jnp.take(x, indices, axis=0), basis.primitives)
        c = jnp.take(basis.coefficients, indices)
        return p, c

    ii, jj = triu_indices(basis.num_primitives)
    lhs, cl = take_primitives(ii.reshape(-1))
    rhs, cr = take_primitives(jj.reshape(-1))
    return (ii, cl, lhs), (jj, cr, rhs)



"""Many electron Hamiltonian with Density Functional Theory or Hartree-Fock."""

from typing import Literal, Optional, Tuple, get_args
from functools import partial

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.numpy.linalg as jnl
import optimistix as optx
from jaxtyping import Array, ScalarLike

from mess.basis import Basis
from mess.integrals import eri_basis, kinetic_basis, nuclear_basis, overlap_basis
from mess.interop import to_pyscf
from mess.mesh import Mesh, density, density_and_grad, xcmesh_from_pyscf
from mess.orthnorm import symmetric
from mess.structure import nuclear_energy
from mess.types import FloatNxN, OrthNormTransform
from mess.xcfunctional import (
    gga_correlation_lyp,
    gga_correlation_pbe,
    gga_exchange_b88,
    gga_exchange_pbe,
    lda_correlation_vwn,
    lda_exchange,
)

xcstr = Literal["lda", "pbe", "pbe0", "b3lyp", "hfx"]
IntegralBackend = Literal["mess", "pyscf_cart", "pyscf_sph"]


class OneElectron(eqx.Module):
    overlap: FloatNxN
    kinetic: FloatNxN
    nuclear: FloatNxN

    def __init__(self, basis: Basis, backend: IntegralBackend = "mess"):
        """_summary_

        Args:
            basis (Basis): _description_
            backend (IntegralBackend, optional): _description_. Defaults to "mess".

        Raises:
            ValueError: _description_
            ValueError: _description_

        Returns:
            _type_: _description_
        """
        if backend == "mess":
            self.overlap = overlap_basis(basis)
            self.kinetic = kinetic_basis(basis)
            self.nuclear = nuclear_basis(basis).sum(axis=0)
        elif backend.startswith("pyscf_"):
            mol = to_pyscf(basis.structure, basis.basis_name)
            kind = backend.split("_")[1]
            S = jnp.array(mol.intor(f"int1e_ovlp_{kind}"))
            N = 1 / jnp.sqrt(jnp.diagonal(S))
            self.overlap = N[:, jnp.newaxis] * N[jnp.newaxis, :] * S
            self.kinetic = jnp.array(mol.intor(f"int1e_kin_{kind}"))
            self.nuclear = jnp.array(mol.intor(f"int1e_nuc_{kind}"))


class TwoElectron(eqx.Module):
    eri: Array

    def __init__(self, basis: Basis, backend: str = "mess"):
        """

        Args:
            basis (Basis): the basis set used to build the electron repulsion integrals
            backend (str, optional): Integral backend used. Defaults to "mess".
        """
        super().__init__()
        if backend == "mess":
            self.eri = eri_basis(basis)
        elif backend.startswith("pyscf_"):
            mol = to_pyscf(basis.structure, basis.basis_name)
            kind = backend.split("_")[1]
            self.eri = jnp.array(mol.intor(f"int2e_{kind}", aosym="s1"))

    def coloumb(self, P: FloatNxN) -> FloatNxN:
        """Build the Coloumb matrix (classical electrostatic) from the density matrix.

        Args:
            P (FloatNxN): the density matrix

        Returns:
            FloatNxN: Coloumb matrix
        """
        return jnp.einsum("kl,ijkl->ij", P, self.eri)

    def exchange(self, P: FloatNxN) -> FloatNxN:
        """Build the quantum-mechanical exchange matrix from the density matrix

        Args:
            P (FloatNxN): the density matrix

        Returns:
            FloatNxN: Exchange matrix
        """
        return jnp.einsum("ij,ikjl->kl", P, self.eri)


class HartreeFockExchange(eqx.Module):
    two_electron: TwoElectron

    def __init__(self, two_electron: TwoElectron):
        self.two_electron = two_electron

    def __call__(self, P: FloatNxN) -> ScalarLike:
        K = self.two_electron.exchange(P)
        return -0.25 * jnp.sum(P * K)


class LDA(eqx.Module):
    basis: Basis
    mesh: Mesh

    def __init__(self, basis: Basis):
        self.basis = basis
        self.mesh = xcmesh_from_pyscf(basis.structure)

    def __call__(self, P: FloatNxN) -> ScalarLike:
        rho = density(self.basis, self.mesh, P)
        eps_xc = lda_exchange(rho) + lda_correlation_vwn(rho)
        E_xc = jnp.einsum("i,i,i", self.mesh.weights, rho, eps_xc)
        return E_xc


class PBE(eqx.Module):
    basis: Basis
    mesh: Mesh

    def __init__(self, basis: Basis):
        self.basis = basis
        self.mesh = xcmesh_from_pyscf(basis.structure)

    def __call__(self, P: FloatNxN) -> ScalarLike:
        rho, grad_rho = density_and_grad(self.basis, self.mesh, P)
        eps_xc = gga_exchange_pbe(rho, grad_rho) + gga_correlation_pbe(rho, grad_rho)
        E_xc = jnp.einsum("i,i,i", self.mesh.weights, rho, eps_xc)
        return E_xc


class PBE0(eqx.Module):
    basis: Basis
    mesh: Mesh
    hfx: HartreeFockExchange

    def __init__(self, basis: Basis, two_electron: TwoElectron):
        self.basis = basis
        self.mesh = xcmesh_from_pyscf(basis.structure)
        self.hfx = HartreeFockExchange(two_electron)

    def __call__(self, P: FloatNxN) -> ScalarLike:
        rho, grad_rho = density_and_grad(self.basis, self.mesh, P)
        e = 0.75 * gga_exchange_pbe(rho, grad_rho) + gga_correlation_pbe(rho, grad_rho)
        E_xc = jnp.einsum("i,i,i", self.mesh.weights, rho, e)
        return E_xc + 0.25 * self.hfx(P)


class B3LYP(eqx.Module):
    basis: Basis
    mesh: Mesh
    hfx: HartreeFockExchange

    def __init__(self, basis: Basis, two_electron: TwoElectron):
        self.basis = basis
        self.mesh = xcmesh_from_pyscf(basis.structure)
        self.hfx = HartreeFockExchange(two_electron)

    def __call__(self, P: FloatNxN) -> ScalarLike:
        rho, grad_rho = density_and_grad(self.basis, self.mesh, P)
        eps_x = 0.08 * lda_exchange(rho) + 0.72 * gga_exchange_b88(rho, grad_rho)
        vwn_c = (1 - 0.81) * lda_correlation_vwn(rho)
        lyp_c = 0.81 * gga_correlation_lyp(rho, grad_rho)
        b3lyp_xc = eps_x + vwn_c + lyp_c
        E_xc = jnp.einsum("i,i,i", self.mesh.weights, rho, b3lyp_xc)
        return E_xc + 0.2 * self.hfx(P)


def build_xcfunc(
    xc_method: xcstr, basis: Basis, two_electron: Optional[TwoElectron] = None
) -> eqx.Module:
    if two_electron is None and xc_method in ("pbe0", "b3lyp"):
        raise ValueError(
            f"Hybrid functional {xc_method} requires providing TwoElectron integrals"
        )

    match xc_method:
        case "lda":
            return LDA(basis)
        case "pbe":
            return PBE(basis)
        case "pbe0":
            return PBE0(basis, two_electron)
        case "b3lyp":
            return B3LYP(basis, two_electron)
        case "hfx":
            return HartreeFockExchange(two_electron)
        case _:
            methods = get_args(xcstr)
            methods = ", ".join(methods)
            msg = f"Unsupported exchange-correlation option: {xc_method}."
            msg += f"\nMust be one of the following: {methods}"
            raise ValueError(msg)


class Hamiltonian(eqx.Module):
    X: FloatNxN
    H_core: FloatNxN
    basis: Basis
    two_electron: TwoElectron
    xcfunc: eqx.Module

    def __init__(
        self,
        basis: Basis,
        ont: OrthNormTransform = symmetric,
        xc_method: xcstr = "lda",
        backend: IntegralBackend = "pyscf_cart",
    ):
        super().__init__()
        self.basis = basis
        one_elec = OneElectron(basis, backend=backend)
        S = one_elec.overlap
        self.X = ont(S)
        self.H_core = one_elec.kinetic + one_elec.nuclear
        self.two_electron = TwoElectron(basis, backend=backend)
        self.xcfunc = build_xcfunc(xc_method, basis, self.two_electron)

    def __call__(self, P: FloatNxN) -> ScalarLike:
        E_core = jnp.sum(self.H_core * P)
        E_xc = self.xcfunc(P)
        J = self.two_electron.coloumb(P)
        E_es = 0.5 * jnp.sum(J * P)
        E = E_core + E_xc + E_es
        return E

    def orthonormalise(self, Z: FloatNxN) -> FloatNxN:
        C = self.X @ jnl.qr(Z).Q
        return C


@partial(jax.jit, static_argnames=("max_steps", "atol", "rtol"))
def minimise(
    H: Hamiltonian,
    max_steps: Optional[int] = None,
    atol: float = 1e-6,
    rtol: float = 1e-5,
) -> Tuple[ScalarLike, FloatNxN, optx.Solution]:
    """Solve for the electronic coefficients that minimise the total energy

    This function takes a Hamiltonian built for a given basis set and molecular
    structure, and finds the electronic coefficients that minimise the total energy.
    The optimisation is performed using the BFGS algorithm.

    Args:
        H (Hamiltonian): The Hamiltonian for a given basis set and molecular structure.
        max_steps (Optional[int]): Maximum number of minimizer steps. Defaults to None.
        atol (float): Absolute tolerance for convergence. Defaults to 1e-6.
        rtol (float): Relative tolerance for convergence. Defaults to 1e-5.

    Returns:
        Tuple[ScalarLike, FloatNxN, optimistix.Solution]: A tuple containing:
            - total energy in atomic units
            - coefficient matrix C that minimizes the Hamiltonian
            - the optimistix.Solution object
    """

    def f(Z, _):
        C = H.orthonormalise(Z)
        P = H.basis.density_matrix(C)
        return H(P)

    solver = optx.BFGS(atol=atol, rtol=rtol)
    Z = jnp.eye(H.basis.num_orbitals)
    sol = optx.minimise(f, solver, Z, max_steps=max_steps)
    C = H.orthonormalise(sol.value)
    P = H.basis.density_matrix(C)
    E_elec = H(P)
    E_total = E_elec + nuclear_energy(H.basis.structure)
    return E_total, C, sol



"""MESS: Modern Electronic Structure Simulations

The mess simulation environment can be customised by setting the following environment
variables:
  * ``MESS_ENABLE_FP64``: enables float64 precision. Defaults to ``True``.
  * ``MESS_CACHE_DIR``: JIT compilation cache location. Defaults to ``~/.cache/mess``
"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("mess")
except PackageNotFoundError:  # Package is not installed
    __version__ = "dev"

from mess.basis import basisset
from mess.hamiltonian import Hamiltonian, minimise
from mess.structure import molecule

__all__ = ["molecule", "Hamiltonian", "minimise", "basisset"]


def parse_bool(value: str) -> bool:
    value = value.lower()

    if value in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif value in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError("Failed to parse {value} as a boolean")


def _setup_env():
    import os
    import os.path as osp

    from jax import config
    from jax.experimental.compilation_cache import compilation_cache as cc

    enable_fp64 = parse_bool(os.environ.get("MESS_ENABLE_FP64", "True"))
    config.update("jax_enable_x64", enable_fp64)

    cache_dir = str(os.environ.get("MESS_CACHE_DIR", osp.expanduser("~/.cache/mess")))
    cc.set_cache_dir(cache_dir)


_setup_env()



import os.path as osp

from jax import config
from jax.experimental.compilation_cache import compilation_cache as cc


def pytest_sessionstart(session):
    cache_dir = osp.expanduser("~/.cache/mess")
    print(f"Initializing JAX compilation cache dir: {cache_dir}")
    cc.set_cache_dir(cache_dir)
    config.update("jax_persistent_cache_min_compile_time_secs", 0.1)


def is_mem_limited():
    # Check if we are running on a limited memory host (e.g. github action)
    import psutil

    total_mem_gib = psutil.virtual_memory().total // 1024**3
    return total_mem_gib < 10



import jax.numpy as jnp
import pytest
from numpy.testing import assert_allclose

from mess.special import (
    binom_beta,
    binom_fori,
    binom_lookup,
    factorial2_fori,
    factorial2_lookup,
    factorial_fori,
    factorial_gamma,
    factorial_lookup,
)


def test_factorial():
    x = jnp.array([1, 2, 3, 4, 5, 6, 7, 8])
    expect = jnp.array([1, 2, 6, 24, 120, 720, 5040, 40320])
    assert_allclose(factorial_fori(x, x[-1]), expect)
    assert_allclose(factorial_lookup(x, x[-1]), expect)
    assert_allclose(factorial_gamma(x), expect)


def test_factorial2():
    x = jnp.array([1, 2, 3, 4, 5, 6, 7, 8])
    expect = jnp.array([1, 2, 3, 8, 15, 48, 105, 384])
    assert_allclose(factorial2_fori(x), expect)
    assert_allclose(factorial2_fori(0), 1)

    assert_allclose(factorial2_lookup(x), expect)
    assert_allclose(factorial2_lookup(0), 1)


@pytest.mark.parametrize("binom_func", [binom_beta, binom_fori, binom_lookup])
def test_binom(binom_func):
    x = jnp.array([4, 4, 4, 4])
    y = jnp.array([1, 2, 3, 4])
    expect = jnp.array([4, 6, 4, 1])
    assert_allclose(binom_func(x, y), expect)

    zero = jnp.array([0])
    assert_allclose(binom_func(zero, y), jnp.zeros_like(x))
    assert_allclose(binom_func(x, zero), jnp.ones_like(y))
    assert_allclose(binom_func(y, y), jnp.ones_like(y))

    one = jnp.array([1])
    assert_allclose(binom_func(one, one), one)
    assert_allclose(binom_func(zero, -one), zero)
    assert_allclose(binom_func(zero, zero), one)



import numpy as np
import pytest
from jax.experimental import enable_x64
from numpy.testing import assert_allclose
from pyscf import dft

from mess.basis import basisset
from mess.hamiltonian import Hamiltonian
from mess.interop import to_pyscf
from mess.structure import Structure

cases = {
    "hfx": "hf,",
    "lda": "slater,vwn_rpa",
    "pbe": "gga_x_pbe,gga_c_pbe",
    "pbe0": "pbe0",
    "b3lyp": "b3lyp",
}


@pytest.mark.parametrize("inputs", cases.items(), ids=cases.keys())
def test_energy(inputs):
    with enable_x64(True):
        xc_method, scfxc = inputs
        mol = Structure(np.asarray(2), np.zeros(3))
        basis_name = "6-31g"
        basis = basisset(mol, basis_name)
        scfmol = to_pyscf(mol, basis_name=basis_name)
        s = dft.RKS(scfmol, xc=scfxc)
        s.kernel()
        P = np.asarray(s.make_rdm1())

        H = Hamiltonian(basis=basis, xc_method=xc_method)
        actual = H(P)
        expect = s.energy_tot()
        assert_allclose(actual, expect, atol=1e-6)



"""Vanilla self-consistent field solver implementation."""

import jax.numpy as jnp
import jax.numpy.linalg as jnl
from jax.lax import while_loop

from mess.basis import Basis
from mess.integrals import eri_basis, kinetic_basis, nuclear_basis, overlap_basis
from mess.structure import nuclear_energy
from mess.orthnorm import cholesky
from mess.types import OrthNormTransform


def scf(
    basis: Basis,
    otransform: OrthNormTransform = cholesky,
    max_iters: int = 32,
    tolerance: float = 1e-4,
):
    """ """
    # init
    Hcore = kinetic_basis(basis) + nuclear_basis(basis).sum(axis=0)
    S = overlap_basis(basis)
    eri = eri_basis(basis)

    # initial guess for MO coeffs
    X = otransform(S)
    C = X @ jnl.eigh(X.T @ Hcore @ X)[1]

    # setup self-consistent iteration as a while loop
    counter = 0
    E = 0.0
    E_prev = 2 * tolerance
    scf_args = (counter, E, E_prev, C)

    def while_cond(scf_args):
        counter, E, E_prev, _ = scf_args
        return (counter < max_iters) & (jnp.abs(E - E_prev) > tolerance)

    def while_body(scf_args):
        counter, E, E_prev, C = scf_args
        P = basis.occupancy * C @ C.T
        J = jnp.einsum("kl,ijkl->ij", P, eri)
        K = jnp.einsum("ij,ikjl->kl", P, eri)
        G = J - 0.5 * K
        H = Hcore + G
        C = X @ jnl.eigh(X.T @ H @ X)[1]
        E_prev = E
        E = 0.5 * jnp.sum(Hcore * P) + 0.5 * jnp.sum(H * P)
        return (counter + 1, E, E_prev, C)

    _, E_electronic, _, _ = while_loop(while_cond, while_body, scf_args)
    E_nuclear = nuclear_energy(basis.structure)
    return E_nuclear + E_electronic



"""Container for molecular structures"""

from typing import List

import equinox as eqx
import numpy as np
import jax.numpy as jnp
from jax import value_and_grad
from periodictable import elements

from mess.types import FloatNx3, IntN
from mess.units import to_bohr


class Structure(eqx.Module):
    atomic_number: IntN
    position: FloatNx3

    def __post_init__(self):
        # single atom case
        self.atomic_number = np.atleast_1d(self.atomic_number)
        self.position = np.atleast_2d(self.position)

    @property
    def num_atoms(self) -> int:
        return len(self.atomic_number)

    @property
    def atomic_symbol(self) -> List[str]:
        return [elements[z].symbol for z in self.atomic_number]

    @property
    def num_electrons(self) -> int:
        return np.sum(self.atomic_number)

    def _repr_html_(self):
        import py3Dmol
        from mess.plot import plot_molecule

        v = py3Dmol.view()
        plot_molecule(v, self)
        return v._repr_html_()


def molecule(name: str) -> Structure:
    """Builds a few sample molecules

    Args:
        name (str): either "h2" or "water". More to be added.

    Raises:
        NotImplementedError: _description_

    Returns:
        Structure: the built molecule as a Structure object
    """

    name = name.lower()

    if name == "h2":
        return Structure(
            atomic_number=np.array([1, 1]),
            position=np.array([[0.0, 0.0, 0.0], [1.4, 0.0, 0.0]]),
        )

    if name == "water":
        r"""Single water molecule
        Structure of single water molecule calculated with DFT using B3LYP
        functional and 6-31+G** basis set <https://cccbdb.nist.gov/>"""
        return Structure(
            atomic_number=np.array([8, 1, 1]),
            position=to_bohr(
                np.array([
                    [0.0000, 0.0000, 0.1165],
                    [0.0000, 0.7694, -0.4661],
                    [0.0000, -0.7694, -0.4661],
                ])
            ),
        )

    raise NotImplementedError(f"No structure registered for: {name}")


def nuclear_energy(structure: Structure) -> float:
    r"""Nuclear electrostatic interaction energy

    Evaluated by taking sum over all unique pairs of atom centers:

    .. math:: \\sum_{j > i} \\frac{Z_i Z_j}{|\\mathbf{r}_i - \\mathbf{r}_j|}

    where :math:`z_i` is the charge of the ith atom (the atomic number).

    Args:
        structure (Structure): input structure

    Returns:
        float: the total nuclear repulsion energy
    """
    idx, jdx = jnp.triu_indices(structure.num_atoms, 1)
    u = structure.atomic_number[idx] * structure.atomic_number[jdx]
    rij = structure.position[idx, :] - structure.position[jdx, :]
    return jnp.sum(u / jnp.linalg.norm(rij, axis=1))


def nuclear_energy_and_force(structure: Structure):
    @value_and_grad
    def energy_and_grad(pos, rest):
        return nuclear_energy(eqx.combine(pos, rest))

    pos, rest = eqx.partition(structure, lambda x: id(x) == id(structure.position))
    E, grad = energy_and_grad(pos, rest)
    return E, -grad.position


def cubic_hydrogen(n: int) -> Structure:
    """
    Builds a Structure of hydrogen atoms arranged in a simple cubic lattice.

    Args:
        n (int): The number of hydrogen atoms for the cubic cell. For example, n=4 will
        build a 4x4x4 cubic lattice.

    Raises:
        ValueError: If n is less than 1.

    Returns:
        Structure: A Structure object representing the cubic lattice of hydrogen atoms.
    """
    if n < 1:
        raise ValueError("Expect at least one hydrogen atom in cubic lattice")

    b = 1.4 * np.arange(0, n)
    pos = np.stack(np.meshgrid(b, b, b)).reshape(3, -1).T
    pos = np.round(pos - np.mean(pos, axis=0), decimals=3)
    return Structure(np.ones(pos.shape[0], dtype=np.int64), pos)
